@@ -90,14 +90,15 @@ void kcapi_dolog(int severity, const char *fmt, ...)
  * Internal logic
  ************************************************************/
 
-int _kcapi_common_accept(struct kcapi_handle *handle, int *fdptr)
+int _kcapi_common_accept(struct kcapi_handle *handle)
 {
+	struct kcapi_handle_tfm *tfm = handle->tfm;
 	int fd;
 
-	if (*fdptr != -1)
+	if (*_kcapi_get_opfd(handle) != -1)
 		return 0;
 
-	fd = accept(handle->tfmfd, NULL, 0);
+	fd = accept(tfm->tfmfd, NULL, 0);
 	if (fd == -1) {
 		int errsv;
 
@@ -107,15 +108,16 @@ int _kcapi_common_accept(struct kcapi_handle *handle, int *fdptr)
 	}
 	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: accept syscall successful");
 
-	*fdptr = fd;
+	*_kcapi_get_opfd(handle) = fd;
 
 	return 0;
 }
 
-int32_t _kcapi_common_send_meta_fd(struct kcapi_handle *handle, int *fdptr,
-				   struct iovec *iov, uint32_t iovlen,
-				   uint32_t enc, uint32_t flags)
+int32_t _kcapi_common_send_meta(struct kcapi_handle *handle,
+				struct iovec *iov, uint32_t iovlen,
+				uint32_t enc, uint32_t flags)
 {
+	struct kcapi_handle_tfm *tfm = handle->tfm;
 	int32_t ret;
 	char buffer_static[80] = { 0 };
 	char *buffer_p = buffer_static;
@@ -129,7 +131,7 @@ int32_t _kcapi_common_send_meta_fd(struct kcapi_handle *handle, int *fdptr,
 	/* IV data */
 	struct af_alg_iv *alg_iv = NULL;
 	uint32_t iv_msg_size = handle->cipher.iv ?
-			  CMSG_SPACE(sizeof(*alg_iv) + handle->info.ivsize) :
+			  CMSG_SPACE(sizeof(*alg_iv) + tfm->info.ivsize) :
 			  0;
 
 	/* AEAD data */
@@ -142,7 +144,7 @@ int32_t _kcapi_common_send_meta_fd(struct kcapi_handle *handle, int *fdptr,
 		iv_msg_size +			/* IV */
 		assoc_msg_size;			/* AEAD associated data size */
 
-	ret = _kcapi_common_accept(handle, fdptr);
+	ret = _kcapi_common_accept(handle);
 	if (ret)
 		return ret;
 
@@ -187,8 +189,8 @@ int32_t _kcapi_common_send_meta_fd(struct kcapi_handle *handle, int *fdptr,
 		header->cmsg_type = ALG_SET_IV;
 		header->cmsg_len = iv_msg_size;
 		alg_iv = (void*)CMSG_DATA(header);
-		alg_iv->ivlen = handle->info.ivsize;
-		memcpy(alg_iv->iv, handle->cipher.iv, handle->info.ivsize);
+		alg_iv->ivlen = tfm->info.ivsize;
+		memcpy(alg_iv->iv, handle->cipher.iv, tfm->info.ivsize);
 	}
 
 	/* set AEAD information */
@@ -206,7 +208,7 @@ int32_t _kcapi_common_send_meta_fd(struct kcapi_handle *handle, int *fdptr,
 		*assoclen = handle->aead.assoclen;
 	}
 
-	ret = sendmsg(*fdptr, &msg, flags);
+	ret = sendmsg(*_kcapi_get_opfd(handle), &msg, flags);
 	if (ret < 0)
 		ret = -errno;
 	kcapi_dolog(KCAPI_LOG_DEBUG,
@@ -219,14 +221,14 @@ out:
 	return ret;
 }
 
-int32_t _kcapi_common_send_data_fd(struct kcapi_handle *handle, int *fdptr,
-				   struct iovec *iov, uint32_t iovlen,
-				   uint32_t flags)
+int32_t _kcapi_common_send_data(struct kcapi_handle *handle,
+				struct iovec *iov, uint32_t iovlen,
+				uint32_t flags)
 {
 	struct msghdr msg;
 	int32_t ret;
 
-	ret = _kcapi_common_accept(handle, fdptr);
+	ret = _kcapi_common_accept(handle);
 	if (ret)
 		return ret;
 
@@ -238,7 +240,7 @@ int32_t _kcapi_common_send_data_fd(struct kcapi_handle *handle, int *fdptr,
 	msg.msg_iov = iov;
 	msg.msg_iovlen = iovlen;
 
-	ret = sendmsg(*fdptr, &msg, flags);
+	ret = sendmsg(*_kcapi_get_opfd(handle), &msg, flags);
 	if (ret < 0)
 		ret = -errno;
 	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: sendmsg syscall returned %d",
@@ -247,9 +249,9 @@ int32_t _kcapi_common_send_data_fd(struct kcapi_handle *handle, int *fdptr,
 	return ret;
 }
 
-int32_t _kcapi_common_vmsplice_iov_fd(struct kcapi_handle *handle, int *fdptr,
-				      struct iovec *iov, unsigned long iovlen,
-				      uint32_t flags)
+int32_t _kcapi_common_vmsplice_iov(struct kcapi_handle *handle,
+				   struct iovec *iov, unsigned long iovlen,
+				   uint32_t flags)
 {
 	int32_t ret = 0;
 	uint32_t inlen = 0;
@@ -269,7 +271,7 @@ int32_t _kcapi_common_vmsplice_iov_fd(struct kcapi_handle *handle, int *fdptr,
 					       (flags & SPLICE_F_MORE) ?
 					        MSG_MORE : 0);
 
-	ret = _kcapi_common_accept(handle, fdptr);
+	ret = _kcapi_common_accept(handle);
 	if (ret)
 		return ret;
 
@@ -289,7 +291,8 @@ int32_t _kcapi_common_vmsplice_iov_fd(struct kcapi_handle *handle, int *fdptr,
 		return -EFAULT;
 	}
 
-	ret = splice(handle->pipes[0], NULL, *fdptr, NULL, ret, flags);
+	ret = splice(handle->pipes[0], NULL, *_kcapi_get_opfd(handle), NULL,
+		     ret, flags);
 	if (ret < 0)
 		ret = -errno;
 	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: splice syscall returned %d", ret);
@@ -297,9 +300,9 @@ int32_t _kcapi_common_vmsplice_iov_fd(struct kcapi_handle *handle, int *fdptr,
 	return ret;
 }
 
-int32_t _kcapi_common_vmsplice_chunk_fd(struct kcapi_handle *handle, int *fdptr,
-				        const uint8_t *in, uint32_t inlen,
-				        uint32_t flags)
+int32_t _kcapi_common_vmsplice_chunk(struct kcapi_handle *handle,
+				     const uint8_t *in, uint32_t inlen,
+				     uint32_t flags)
 {
 	struct iovec iov;
 	uint32_t processed = 0;
@@ -312,7 +315,7 @@ int32_t _kcapi_common_vmsplice_chunk_fd(struct kcapi_handle *handle, int *fdptr,
 	if (!inlen)
 		return _kcapi_common_send_data(handle, NULL, 0, sflags);
 
-	ret = _kcapi_common_accept(handle, fdptr);
+	ret = _kcapi_common_accept(handle);
 	if (ret)
 		return ret;
 
@@ -338,7 +341,8 @@ int32_t _kcapi_common_vmsplice_chunk_fd(struct kcapi_handle *handle, int *fdptr,
 				    "AF_ALG: vmsplice syscall returned %d",
 				    ret);
 
-			ret = splice(handle->pipes[0], NULL, *fdptr, NULL, ret,
+			ret = splice(handle->pipes[0], NULL,
+				     *_kcapi_get_opfd(handle), NULL, ret,
 				     flags);
 			if (ret < 0) {
 				ret = -errno;
@@ -437,7 +441,7 @@ int _kcapi_aio_send_iov(struct kcapi_handle *handle, struct iovec *iov,
 	return 0;
 }
 
-int _kcapi_aio_read_iov_fd(struct kcapi_handle *handle, int *fdptr,
+int _kcapi_aio_read_iov(struct kcapi_handle *handle,
 			   struct iovec *iov, uint32_t iovlen)
 {
 	struct iocb *cb = handle->aio.cio;
@@ -459,7 +463,7 @@ int _kcapi_aio_read_iov_fd(struct kcapi_handle *handle, int *fdptr,
 		}
 
 		memset(cb, 0, sizeof(*cb));
-		cb->aio_fildes = *fdptr;
+		cb->aio_fildes = *_kcapi_get_opfd(handle);
 		cb->aio_lio_opcode = IOCB_CMD_PREAD;
 		cb->aio_buf = (unsigned long)iov->iov_base;
 		cb->aio_offset = 0;
@@ -490,13 +494,13 @@ int _kcapi_aio_read_iov_fd(struct kcapi_handle *handle, int *fdptr,
 	return _kcapi_aio_read_all(handle, iovlen, NULL);
 }
 
-int32_t _kcapi_common_recv_data_fd(struct kcapi_handle *handle, int *fdptr,
-				   struct iovec *iov, uint32_t iovlen)
+int32_t _kcapi_common_recv_data(struct kcapi_handle *handle,
+				struct iovec *iov, uint32_t iovlen)
 {
 	struct msghdr msg;
 	int32_t ret;
 
-	ret = _kcapi_common_accept(handle, fdptr);
+	ret = _kcapi_common_accept(handle);
 	if (ret)
 		return ret;
 
@@ -508,7 +512,7 @@ int32_t _kcapi_common_recv_data_fd(struct kcapi_handle *handle, int *fdptr,
 	msg.msg_iov = iov;
 	msg.msg_iovlen = iovlen;
 
-	ret = recvmsg(*fdptr, &msg, 0);
+	ret = recvmsg(*_kcapi_get_opfd(handle), &msg, 0);
 	if (ret < 0)
 		ret = -errno;
 	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: recvmsg syscall returned %d",
@@ -538,32 +542,46 @@ int32_t _kcapi_common_recv_data_fd(struct kcapi_handle *handle, int *fdptr,
 	return ret;
 }
 
-int32_t _kcapi_common_read_data_fd(struct kcapi_handle *handle, int *fdptr,
+int32_t _kcapi_common_read_data(struct kcapi_handle *handle,
 				   uint8_t *out, uint32_t outlen)
 {
-	int32_t ret;
+	int ret;
+	int32_t totallen = 0;
 
 	if (outlen > INT_MAX)
 		return -EMSGSIZE;
 
-	ret = _kcapi_common_accept(handle, fdptr);
+	ret = _kcapi_common_accept(handle);
 	if (ret)
 		return ret;
 
-	ret = read(*fdptr, out, outlen);
-	if (ret < 0)
-		ret = -errno;
-	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: read syscall returned %d", ret);
 
-	return ret;
+	if (outlen) {
+		do {
+			ret = read(*_kcapi_get_opfd(handle), out, outlen);
+			if (ret > 0) {
+				out += ret;
+				outlen -= ret;
+				totallen += ret;
+			}
+			kcapi_dolog(KCAPI_LOG_DEBUG,
+				    "AF_ALG: read syscall returned %d", ret);
+		} while ((ret > 0 || errno == EINTR) && outlen);
+
+		if (ret < 0)
+			return -errno;
+	}
+
+	return totallen;
 }
 
 int _kcapi_common_setkey(struct kcapi_handle *handle,
 			 const uint8_t *key, uint32_t keylen)
 {
+	struct kcapi_handle_tfm *tfm = handle->tfm;
 	int ret;
 
-	ret = setsockopt(handle->tfmfd, SOL_ALG, ALG_SET_KEY, key, keylen);
+	ret = setsockopt(tfm->tfmfd, SOL_ALG, ALG_SET_KEY, key, keylen);
 	if (ret < 0)
 		ret = -errno;
 	kcapi_dolog(KCAPI_LOG_DEBUG,
@@ -576,6 +594,7 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 				  const char *ciphername,
 				  int drivername)
 {
+	struct kcapi_handle_tfm *tfm = handle->tfm;
 	int ret = -EFAULT;
 
 	/* NETLINK_CRYPTO specific */
@@ -733,41 +752,51 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 		struct rtattr *rta = tb[CRYPTOCFGA_REPORT_HASH];
 		struct crypto_report_hash *rsh =
 			(struct crypto_report_hash *) RTA_DATA(rta);
-		handle->info.hash_digestsize = rsh->digestsize;
-		handle->info.blocksize = rsh->blocksize;
-		kcapi_dolog(KCAPI_LOG_DEBUG, "Get cipher info: hash with digestsize %u,  blocksize %u",
+		tfm->info.hash_digestsize = rsh->digestsize;
+		tfm->info.blocksize = rsh->blocksize;
+		kcapi_dolog(KCAPI_LOG_DEBUG,
+			    "Get cipher info: hash with digestsize %u,  blocksize %u",
 			    rsh->digestsize, rsh->blocksize);
+
+		tfm->info.cipher_type = KCAPI_CIPHER_MD;
 	}
 	if (tb[CRYPTOCFGA_REPORT_BLKCIPHER]) {
 		struct rtattr *rta = tb[CRYPTOCFGA_REPORT_BLKCIPHER];
 		struct crypto_report_blkcipher *rblk =
 			(struct crypto_report_blkcipher *) RTA_DATA(rta);
-		handle->info.blocksize = rblk->blocksize;
-		handle->info.ivsize = rblk->ivsize;
-		handle->info.blk_min_keysize = rblk->min_keysize;
-		handle->info.blk_max_keysize = rblk->max_keysize;
+		tfm->info.blocksize = rblk->blocksize;
+		tfm->info.ivsize = rblk->ivsize;
+		tfm->info.blk_min_keysize = rblk->min_keysize;
+		tfm->info.blk_max_keysize = rblk->max_keysize;
 		kcapi_dolog(KCAPI_LOG_DEBUG, "Get cipher info: block cipher with blocksize %u, ivsize %u, minimum keysize %u, maximum keysize %u",
 			    rblk->blocksize, rblk->ivsize, rblk->min_keysize,
 			    rblk->max_keysize);
+
+		tfm->info.cipher_type = KCAPI_CIPHER_SKCIPHER;
 	}
 	if (tb[CRYPTOCFGA_REPORT_AEAD]) {
 		struct rtattr *rta = tb[CRYPTOCFGA_REPORT_AEAD];
 		struct crypto_report_aead *raead =
 			(struct crypto_report_aead *) RTA_DATA(rta);
-		handle->info.blocksize = raead->blocksize;
-		handle->info.ivsize = raead->ivsize;
-		handle->info.aead_maxauthsize = raead->maxauthsize;
+		tfm->info.blocksize = raead->blocksize;
+		tfm->info.ivsize = raead->ivsize;
+		tfm->info.aead_maxauthsize = raead->maxauthsize;
 		kcapi_dolog(KCAPI_LOG_DEBUG, "Get cipher info: AEAD block cipher with blocksize %u, ivsize %u, maximum authentication size %u",
 			    raead->blocksize, raead->ivsize,
 			    raead->maxauthsize);
+
+		tfm->info.cipher_type = KCAPI_CIPHER_AEAD;
 	}
 	if (tb[CRYPTOCFGA_REPORT_RNG]) {
 		struct rtattr *rta = tb[CRYPTOCFGA_REPORT_RNG];
 		struct crypto_report_rng *rrng =
 			(struct crypto_report_rng *) RTA_DATA(rta);
-		handle->info.rng_seedsize = rrng->seedsize;
-		kcapi_dolog(KCAPI_LOG_DEBUG, "Get cipher info: RNG cipher with seedsize %u",
+		tfm->info.rng_seedsize = rrng->seedsize;
+		kcapi_dolog(KCAPI_LOG_DEBUG,
+			    "Get cipher info: RNG cipher with seedsize %u",
 			    rrng->seedsize);
+
+		tfm->info.cipher_type = KCAPI_CIPHER_RNG;
 	}
 	if (tb[CRYPTOCFGA_UNSPEC])
 		kcapi_dolog(KCAPI_LOG_DEBUG,
@@ -784,12 +813,18 @@ static int __kcapi_common_getinfo(struct kcapi_handle *handle,
 	if (tb[CRYPTOCFGA_REPORT_CIPHER])
 		kcapi_dolog(KCAPI_LOG_DEBUG,
 			    "Get cipher info: simple cipher algorithm type received");
-	if (tb[CRYPTOCFGA_REPORT_AKCIPHER])
+	if (tb[CRYPTOCFGA_REPORT_AKCIPHER]) {
 		kcapi_dolog(KCAPI_LOG_DEBUG,
 			    "Get cipher info: asymmetric cipher algorithm type received");
-	if (tb[CRYPTOCFGA_REPORT_KPP])
+
+		tfm->info.cipher_type = KCAPI_CIPHER_AKCIPHER;
+	}
+	if (tb[CRYPTOCFGA_REPORT_KPP]) {
 		kcapi_dolog(KCAPI_LOG_DEBUG,
 			    "Get cipher info: kpp cipher algorithm type received");
+
+		tfm->info.cipher_type = KCAPI_CIPHER_KPP;
+	}
 	if (tb[CRYPTOCFGA_REPORT_ACOMP])
 		kcapi_dolog(KCAPI_LOG_DEBUG,
 			    "Get cipher info: asymmetric compression algorithm type received");
@@ -832,30 +867,41 @@ static inline void _kcapi_aio_destroy(struct kcapi_handle *handle)
 	handle->aio.iocb_ret = NULL;
 }
 
-void _kcapi_handle_destroy_nofree(struct kcapi_handle *handle)
+static void _kcapi_handle_destroy_tfm(struct kcapi_handle *handle)
+{
+	struct kcapi_handle_tfm *tfm;
+	if (!handle || !handle->tfm)
+		return;
+
+	tfm = handle->tfm;
+	if (atomic_dec_and_test(&tfm->refcnt)) {
+		if (tfm->tfmfd != -1)
+			close(tfm->tfmfd);
+		kcapi_memset_secure(tfm, 0, sizeof(*tfm));
+		free(tfm);
+		handle->tfm = NULL;
+	}
+}
+
+void _kcapi_handle_destroy(struct kcapi_handle *handle)
 {
 	if (!handle)
 		return;
-	if (handle->tfmfd != -1)
-		close(handle->tfmfd);
-	if (handle->opfd != -1)
-		close(handle->opfd);
+	if (*_kcapi_get_opfd(handle) != -1)
+		close(*_kcapi_get_opfd(handle));
 	if (handle->pipes[0] != -1)
 		close(handle->pipes[0]);
 	if (handle->pipes[1] != -1)
 		close(handle->pipes[1]);
 	_kcapi_aio_destroy(handle);
+	_kcapi_handle_destroy_tfm(handle);
 	kcapi_memset_secure(handle, 0, sizeof(struct kcapi_handle));
-}
-
-void _kcapi_handle_destroy(struct kcapi_handle *handle)
-{
-	_kcapi_handle_destroy_nofree(handle);
 	free(handle);
 }
 
 static int _kcapi_get_kernver(struct kcapi_handle *handle)
 {
+	struct kcapi_handle_tfm *tfm = handle->tfm;
 	struct utsname kernel;
 	char *saveptr = NULL;
 	char *res = NULL;
@@ -869,19 +915,19 @@ static int _kcapi_get_kernver(struct kcapi_handle *handle)
 		printf("Could not parse kernel version");
 		return -EFAULT;
 	}
-	handle->sysinfo.kernel_maj = strtoul(res, NULL, 10);
+	tfm->sysinfo.kernel_maj = strtoul(res, NULL, 10);
 	res = strtok_r(NULL, ".", &saveptr);
 	if (!res) {
 		printf("Could not parse kernel version");
 		return -EFAULT;
 	}
-	handle->sysinfo.kernel_minor = strtoul(res, NULL, 10);
+	tfm->sysinfo.kernel_minor = strtoul(res, NULL, 10);
 	res = strtok_r(NULL, ".", &saveptr);
 	if (!res) {
 		printf("Could not parse kernel version");
 		return -EFAULT;
 	}
-	handle->sysinfo.kernel_patchlevel = strtoul(res, NULL, 10);
+	tfm->sysinfo.kernel_patchlevel = strtoul(res, NULL, 10);
 
 	return 0;
 }
@@ -890,32 +936,35 @@ static int _kcapi_get_kernver(struct kcapi_handle *handle)
 static bool _kcapi_kernver_ge(struct kcapi_handle *handle, unsigned int maj,
 			      unsigned int minor, unsigned int patchlevel)
 {
-	if (maj < handle->sysinfo.kernel_maj)
+	struct kcapi_handle_tfm *tfm = handle->tfm;
+
+	if (maj < tfm->sysinfo.kernel_maj)
 		return true;
-	if (maj == handle->sysinfo.kernel_maj) {
-		if (minor < handle->sysinfo.kernel_minor)
+	if (maj == tfm->sysinfo.kernel_maj) {
+		if (minor < tfm->sysinfo.kernel_minor)
 			return true;
-		if (minor == handle->sysinfo.kernel_minor) {
-			if (patchlevel <= handle->sysinfo.kernel_patchlevel)
+		if (minor == tfm->sysinfo.kernel_minor) {
+			if (patchlevel <= tfm->sysinfo.kernel_patchlevel)
 				return true;
 		}
 	}
 	return false;
 }
 
-static int _kcapi_aio_init(struct kcapi_handle *handle, const char *type)
+static int _kcapi_aio_init(struct kcapi_handle *handle)
 {
+	struct kcapi_handle_tfm *tfm = handle->tfm;
 	uint32_t i;
 	int err;
 
-	if (!strncmp("aead", type, 4)) {
+	if (tfm->info.cipher_type == KCAPI_CIPHER_AEAD) {
 		if (!_kcapi_kernver_ge(handle, 4, 7, 0)) {
 			kcapi_dolog(KCAPI_LOG_VERBOSE, "AIO support for AEAD cipher not present on current kernel");
 			err = -EOPNOTSUPP;
 			goto err;
 		}
 	}
-	if (!strncmp("skcipher", type, 8)) {
+	if (tfm->info.cipher_type == KCAPI_CIPHER_SKCIPHER) {
 		if (!_kcapi_kernver_ge(handle, 4, 1, 0)) {
 			kcapi_dolog(KCAPI_LOG_VERBOSE, "AIO support for symmetric ciphers not present on current kernel");
 			err = -EOPNOTSUPP;
@@ -976,59 +1025,23 @@ err:
 
 static void _kcapi_handle_flags(struct kcapi_handle *handle)
 {
-	/* new memory structure for AF_ALG AEAD interface */
-	handle->flags.newtag = _kcapi_kernver_ge(handle, 4, 9, 0);
+	/* new memory structure for AF_ALG AEAD interface + fixed hashing empty input */
+	handle->flags.ge_v4_9 = _kcapi_kernver_ge(handle, 4, 9, 0);
 
 	/* older interfaces only processed 16 pages in a row */
 	handle->flags.alg_max_pages = _kcapi_kernver_ge(handle, 4, 11, 0) ?
 				      UINT_MAX : ALG_MAX_PAGES;
 }
 
-int _kcapi_allocated_handle_init(struct kcapi_handle *handle, const char *type,
-				 const char *ciphername, uint32_t flags)
+static int _kcapi_handle_init_op(struct kcapi_handle *handle, uint32_t flags)
 {
-	struct sockaddr_alg sa;
 	int ret;
-	char versionbuffer[50];
 
-	kcapi_versionstring(versionbuffer, sizeof(versionbuffer));
-	kcapi_dolog(KCAPI_LOG_VERBOSE,
-		    "%s - initializing cipher operation with kernel",
-		    versionbuffer);
-
-	handle->opfd = -1;
-	handle->tfmfd = -1;
+	*_kcapi_get_opfd(handle) = -1;
 	handle->pipes[0] = -1;
 	handle->pipes[1] = -1;
 	handle->aio.efd = -1;
 
-	ret = _kcapi_get_kernver(handle);
-	if (ret)
-		return ret;
-
-	memset(&sa, 0, sizeof(sa));
-	sa.salg_family = AF_ALG;
-	snprintf((char *)sa.salg_type, sizeof(sa.salg_type),"%s", type);
-	snprintf((char *)sa.salg_name, sizeof(sa.salg_name),"%s", ciphername);
-
-	handle->tfmfd = socket(AF_ALG, SOCK_SEQPACKET, 0);
-	if (handle->tfmfd == -1) {
-		ret = -errno;
-		kcapi_dolog(KCAPI_LOG_ERR,
-			    "AF_ALG: socket syscall failed (errno: %d)",
-			    ret);
-		return ret;
-	}
-	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: socket syscall passed");
-
-	if (bind(handle->tfmfd, (struct sockaddr *)&sa, sizeof(sa)) == -1) {
-		ret = -errno;
-		kcapi_dolog(KCAPI_LOG_ERR, "AF_ALG: bind failed (errno: %d)",
-			    ret);
-		return ret;
-	}
-	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: bind syscall passed");
-
 	ret = pipe(handle->pipes);
 	if (ret) {
 		ret = -errno;
@@ -1039,16 +1052,8 @@ int _kcapi_allocated_handle_init(struct kcapi_handle *handle, const char *type,
 	}
 	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: pipe syscall passed");
 
-	ret = _kcapi_common_getinfo(handle, ciphername);
-	if (ret) {
-		ret = -errno;
-		kcapi_dolog(KCAPI_LOG_ERR, "NETLINK_CRYPTO: cannot obtain cipher information for %s (is required crypto_user.c patch missing? see documentation)",
-			    ciphername);
-		return ret;
-	}
-
 	if (flags & KCAPI_INIT_AIO) {
-		ret = _kcapi_aio_init(handle, type);
+		ret = _kcapi_aio_init(handle);
 
 		/*
 		 * We complain about kernels without AIO support, but allow
@@ -1065,15 +1070,106 @@ int _kcapi_allocated_handle_init(struct kcapi_handle *handle, const char *type,
 
 	_kcapi_handle_flags(handle);
 
-	kcapi_dolog(KCAPI_LOG_VERBOSE,
-		    "communication for %s with kernel initialized",
-		    ciphername);
-
 	return ret;
 }
 
+static int _kcapi_handle_init_tfm(struct kcapi_handle *handle, const char *type,
+				  const char *ciphername)
+{
+	struct kcapi_handle_tfm *tfm = handle->tfm;
+	struct sockaddr_alg sa;
+	int ret;
+	char versionbuffer[50];
+
+	kcapi_versionstring(versionbuffer, sizeof(versionbuffer));
+	kcapi_dolog(KCAPI_LOG_VERBOSE,
+		    "%s - initializing cipher operation with kernel",
+		    versionbuffer);
+
+	tfm->tfmfd = -1;
+	atomic_set(1, &tfm->refcnt);
+
+	ret = _kcapi_get_kernver(handle);
+	if (ret)
+		return ret;
+
+	memset(&sa, 0, sizeof(sa));
+	sa.salg_family = AF_ALG;
+	snprintf((char *)sa.salg_type, sizeof(sa.salg_type),"%s", type);
+	snprintf((char *)sa.salg_name, sizeof(sa.salg_name),"%s", ciphername);
+
+	tfm->tfmfd = socket(AF_ALG, SOCK_SEQPACKET, 0);
+	if (tfm->tfmfd == -1) {
+		ret = -errno;
+		kcapi_dolog(KCAPI_LOG_ERR,
+			    "AF_ALG: socket syscall failed (errno: %d)",
+			    ret);
+		return ret;
+	}
+	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: socket syscall passed");
+
+	if (bind(tfm->tfmfd, (struct sockaddr *)&sa, sizeof(sa)) == -1) {
+		ret = -errno;
+		kcapi_dolog(KCAPI_LOG_ERR, "AF_ALG: bind failed (errno: %d)",
+			    ret);
+		return ret;
+	}
+	kcapi_dolog(KCAPI_LOG_DEBUG, "AF_ALG: bind syscall passed");
+
+	ret = _kcapi_common_getinfo(handle, ciphername);
+	if (ret) {
+		ret = -errno;
+		kcapi_dolog(KCAPI_LOG_ERR, "NETLINK_CRYPTO: cannot obtain cipher information for %s (is required crypto_user.c patch missing? see documentation)",
+			    ciphername);
+		return ret;
+	}
+
+	return 0;
+}
+
 int _kcapi_handle_init(struct kcapi_handle **caller, const char *type,
 		       const char *ciphername, uint32_t flags)
+{
+	struct kcapi_handle *handle;
+	struct kcapi_handle_tfm *tfm;
+	int ret;
+
+	handle = calloc(1, sizeof(struct kcapi_handle));
+	if (!handle)
+		return -ENOMEM;
+
+	tfm = calloc(1, sizeof(struct kcapi_handle_tfm));
+	if (!tfm) {
+		free(handle);
+		return -ENOMEM;
+	}
+
+	handle->tfm = tfm;
+
+	ret = _kcapi_handle_init_tfm(handle, type, ciphername);
+	if (ret)
+		goto err;
+
+	ret = _kcapi_handle_init_op(handle, flags);
+	if (ret)
+		goto err;
+
+	*caller = handle;
+
+	kcapi_dolog(KCAPI_LOG_VERBOSE,
+		    "communication for %s with kernel initialized",
+		    ciphername);
+
+	return 0;
+
+err:
+	_kcapi_handle_destroy(handle);
+	return ret;
+}
+
+DSO_PUBLIC
+int kcapi_handle_reinit(struct kcapi_handle **newhandle,
+			struct kcapi_handle *existing, uint32_t flags)
 {
 	struct kcapi_handle *handle;
 	int ret;
@@ -1082,12 +1178,27 @@ int _kcapi_handle_init(struct kcapi_handle **caller, const char *type,
 	if (!handle)
 		return -ENOMEM;
 
-	ret = _kcapi_allocated_handle_init(handle, type, ciphername, flags);
-	if (ret)
-		_kcapi_handle_destroy(handle);
-	else
-		*caller = handle;
+	if (!existing || !existing->tfm) {
+		ret = -EINVAL;
+		goto err;
+	}
 
+	atomic_inc(&existing->tfm->refcnt);
+	handle->tfm = existing->tfm;
+
+	ret = _kcapi_handle_init_op(handle, flags);
+	if (ret)
+		goto err;
+
+	*newhandle = handle;
+
+	kcapi_dolog(KCAPI_LOG_VERBOSE,
+		    "new cipher handle from existing handle initialized");
+
+	return 0;
+
+err:
+	_kcapi_handle_destroy(handle);
 	return ret;
 }
 
@@ -1132,14 +1243,14 @@ int32_t _kcapi_cipher_crypt_chunk(struct kcapi_handle *handle,
 {
 	int32_t totallen = 0;
 	uint32_t maxprocess = sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES;
+	int32_t ret;
 
 	if (outlen > INT_MAX)
 		return -EMSGSIZE;
 
-	while (inlen) {
+	while (inlen && outlen) {
 		uint32_t inprocess = inlen;
 		uint32_t outprocess = outlen;
-		int32_t ret = 0;
 
 		/*
 		 * We do not check that sysconf(_SC_PAGESIZE) * ALG_MAX_PAGES is
